(* ************************************************************************* *)
(*                             TinyQCMl                                      *)
(*                   (c) Franois Pessaux - 2014                             *)
(* TinyQCMl is freely distribuable (sources and binaries if some).           *)
(* ************************************************************************* *)

(* Heavily "inspired" from the FoCaLiZe parser. With the permission of the
    authors... I'm one of them ^_^ *)
{
open Lexing ;;
open Parser ;;

type error =
   | Comment_in_string
   | Comment_in_uniline_comment
   | Illegal_character of char
   | Illegal_escape of string
   | Uninitiated_comment
   | Unterminated_comment
   | Unterminated_string
   | Ill_formed_float
;;

exception Error of error * Lexing.position * Lexing.position ;;


let string_of_lex_error = function
  | Comment_in_string ->
      "Non escaped comment separator in string constant"
  | Comment_in_uniline_comment ->
      "Non escaped comment separator in uniline comment"
  | Illegal_character c ->
      "Illegal character (" ^ Char.escaped c ^ ")"
  | Illegal_escape s ->
      "Illegal backslash escape in string or character (" ^ s ^ ")"
  | Uninitiated_comment ->
      "Comment has not started"
  | Unterminated_comment ->
      "Comment not terminated"
  | Unterminated_string ->
      "String literal not terminated"
  | Ill_formed_float ->
      "Ill-formed float literal"
;;



(** {b Descr}: Lexing the string tokens. *)
let initial_string_buffer = String.create 256 ;;
let string_buff = ref initial_string_buffer
and string_index = ref 0 ;;

let reset_string_buffer () =
  string_buff := initial_string_buffer ;
  string_index := 0
;;

let store_string_char c =
  if !string_index >= Bytes.length !string_buff then
    begin
      let new_buff = Bytes.create (Bytes.length !string_buff * 2) in
      Bytes.blit !string_buff 0 new_buff 0 (Bytes.length !string_buff) ;
      string_buff := new_buff
    end ;
  Bytes.unsafe_set !string_buff !string_index c ;
  incr string_index
;;

let get_stored_string () =
  let s = Bytes.to_string (Bytes.sub !string_buff 0 !string_index) in
  string_buff := initial_string_buffer ;
  s
;;


let string_start_pos = ref None ;;
let comment_start_pos = ref [] ;;



(** {b Descr}: Decoding characters} *)
let char_for_character = function
  | 'n' -> '\010'
  | 'r' -> '\013'
  | 'b' -> '\008'
  | 't' -> '\009'
  | c -> c
;;

let char_for_decimal_code lexbuf i =
  let c =
    100 * (Char.code(Lexing.lexeme_char lexbuf i) - 48) +
     10 * (Char.code(Lexing.lexeme_char lexbuf (i + 1)) - 48) +
          (Char.code(Lexing.lexeme_char lexbuf (i + 2)) - 48) in
  if c >= 0 && c <= 255 then Char.chr c else
    raise
      (Error (Illegal_escape (Lexing.lexeme lexbuf),
              lexbuf.lex_start_p,
              lexbuf.lex_curr_p))
;;

let char_for_hexadecimal_code lexbuf i =
  let d1 = Char.code (Lexing.lexeme_char lexbuf i) in
  let val1 =
    if d1 >= 97 then d1 - 87 else
    if d1 >= 65 then d1 - 55 else
    d1 - 48 in
  let d2 = Char.code (Lexing.lexeme_char lexbuf (i + 1)) in
  let val2 =
    if d2 >= 97 then d2 - 87 else
    if d2 >= 65 then d2 - 55 else
    d2 - 48 in
  Char.chr (val1 * 16 + val2)
;;



(** {b Descr}: Keeping the internal buffer locations up to date} *)
let update_loc lexbuf fname line absolute chars =
  let pos = lexbuf.lex_curr_p in
  let new_fname =
    match fname with
    | None -> pos.pos_fname
    | Some s -> s in
  lexbuf.lex_curr_p <- {
    pos with
    pos_fname = new_fname;
    pos_lnum = if absolute then line else pos.pos_lnum + line;
    pos_bol = pos.pos_cnum - chars;
  }
;;

(** Unconditionally set the location as being on line [num],
    in file [fname]. *)
let set_loc lexbuf fname num =
  update_loc lexbuf fname (int_of_string num) true 0
;;

(** Add one to the current line counter of the file being lexed. *)
let incr_line_num lexbuf =
  update_loc lexbuf None 1 false 0
;;

(** Add one to the current line counter of the file being lexed,
    knowing that [spaces] chars are skipped at the beginning of the next
    line. *)
let incr_escaped_line_num lexbuf spaces =
  update_loc lexbuf None 1 false (String.length spaces)
;;

}

(** {3 The main lexer} *)

(** {6 Classifying characters} *)
let newline = ( '\n' | '\r' | "\r\n" )
(** ASCII 010 is linefeed or ['\n'], ASCII 013 is carriage return or ['\r']. *)
let blank = [ '\032' '\009' '\012' ]
(** ASCII 32 is space, ASCII 9 is tab or ['\t'], ASCII 12 is formfeed *)
let white = [ ' ' '\t' ]
let whites = white*
(** Any number of space and tabs (including 0). *)

let space_separator = newline | blank

(** {b Descr}: Classification of characters for numbers. *)
let decimal_digit = [ '0'-'9' ]
let hexadecimal_digit = [ '0'-'9' 'a'-'f' 'A'-'F' ]

(** {Classification of characters for string and character tokens. *)
let escaped_decimal_code = "\\" decimal_digit decimal_digit decimal_digit

let escaped_hexadecimal_code = "\\" 'x' hexadecimal_digit hexadecimal_digit

let escaped_character =
    "\\" [ '(' '[' '{' '-' '\\' '`' '\'' '\"' 'n' 't' 'b' 'r' ' ' '*' '}' ']' ')' ]
    (* ` Helping emacs }]) *)

(** {Definition of float literals} *)

let unsigned_decimal_literal = decimal_digit decimal_digit*
let float_literal =
  unsigned_decimal_literal
  ('.' unsigned_decimal_literal* )?


(** {b Descr}: The main lexer. *)
rule token = parse
  | newline
    { incr_line_num lexbuf;
      token lexbuf }
  | blank +
    { token lexbuf }
  (* Floats. *)
  | float_literal
      { try FLOAT (float_of_string (Lexing.lexeme lexbuf)) with
      | Failure "float_of_string" ->
          raise (Error (Ill_formed_float, lexbuf.lex_start_p,
                 lexbuf.lex_curr_p))
      }
  (* Strings. *)
  | "\""
    { reset_string_buffer ();
      string_start_pos :=
        Some (lexbuf.lex_start_p, lexbuf.lex_curr_p) ;
      string lexbuf;
      begin match !string_start_pos with
      | Some (start_pos, _) -> lexbuf.lex_start_p <- start_pos
      | _ -> assert false end;
      STRING (get_stored_string ()) }
  (* Comments *)
  | "(*"
    { comment_start_pos := [ lexbuf.lex_start_p, lexbuf.lex_curr_p ] ;
      comment lexbuf ;
      token lexbuf }
  | "*)"
    { raise
        (Error (Uninitiated_comment,
                lexbuf.lex_start_p,
                lexbuf.lex_curr_p)) }
  | "--"
    { uniline_comment lexbuf;
      token lexbuf }

  (* Usual simple tokens *)
  | "KO"        { KO }
  | "OK"        { OK }

  (* Punctuation. *)
  | '{' { LBRACE }
  | '}' { RBRACE }
  | '=' { EQUAL }
  | ',' { COMMA }
  | ';' { SEMI }

  | eof { EOF }
  | _
    { raise
        (Error
           (Illegal_character
              (Lexing.lexeme_char lexbuf 0),
              lexbuf.lex_start_p,
              lexbuf.lex_curr_p)) }

  (* Special sub lexer for uni-line comments *)
and uniline_comment = parse
  | ( "(*" | "*)" )
    { raise
        (Error
          (Comment_in_uniline_comment,
           lexbuf.lex_start_p,
           lexbuf.lex_curr_p)) }
  | '\\' newline (whites as spaces)
    { incr_escaped_line_num lexbuf spaces ;
      uniline_comment lexbuf }
  | newline
    { incr_line_num lexbuf }
  | eof
    { raise
        (Error
          (Unterminated_comment,
           lexbuf.lex_start_p,
           lexbuf.lex_curr_p)) }
  | _
    { uniline_comment lexbuf }

  (* Special sub lexer for multi lines possibly nested comments *)
and comment = parse
  | "(*"
    { comment_start_pos :=
        (lexbuf.lex_start_p, lexbuf.lex_curr_p) :: !comment_start_pos ;
      comment lexbuf }
  | '\\' '*'
    { comment lexbuf }
  | "*)"
    { match !comment_start_pos with
      | [] -> assert false
      | [ _ ] -> comment_start_pos := []
      | _ :: l ->
        comment_start_pos := l ;
        comment lexbuf }
  | '\\' newline (whites as spaces)
    { incr_escaped_line_num lexbuf spaces;
      comment lexbuf }
  | newline
    { incr_line_num lexbuf ;
      comment lexbuf }
  | eof
    { match !comment_start_pos with
      | [] -> assert false
      | (start_pos, end_pos) :: _ ->
        comment_start_pos := [] ;
        raise (Error (Unterminated_comment, start_pos, end_pos)) }
  | "--"
    { uniline_comment lexbuf ;
      comment lexbuf }
  | _
    { comment lexbuf }

  (* Special sub lexer for string lexems *)
and string = parse
  | '\"'
    { () }
  | '\\' newline (whites as spaces)
    { incr_escaped_line_num lexbuf spaces ;
      string lexbuf }
  | escaped_character
    (* ` Helping emacs *)
    { store_string_char (char_for_character (Lexing.lexeme_char lexbuf 1)) ;
      string lexbuf }
  | escaped_decimal_code
    { store_string_char (char_for_decimal_code lexbuf 1) ;
      string lexbuf }
  | escaped_hexadecimal_code
    { store_string_char (char_for_hexadecimal_code lexbuf 2) ;
      string lexbuf }
  | "\\%"
    { store_string_char '%' ;
      string lexbuf }
  | '\\' _
    { raise
        (Error
           (Illegal_escape (Lexing.lexeme lexbuf),
           lexbuf.lex_start_p,
           lexbuf.lex_curr_p)) }
  | ( "(*" | "*)" "--" )
    { raise
        (Error
           (Comment_in_string,
            lexbuf.lex_start_p,
            lexbuf.lex_curr_p)) }
  | newline
    { incr_line_num lexbuf;
      match !string_start_pos with
      | Some (start_pos, end_pos) ->
        raise (Error (Unterminated_string, start_pos, end_pos))
      | _ -> assert false }
  | eof
    { match !string_start_pos with
      | Some (start_pos, end_pos) ->
        raise (Error (Unterminated_string, start_pos, end_pos))
      | _ -> assert false }
  | _
    { store_string_char (Lexing.lexeme_char lexbuf 0) ;
      string lexbuf }
