(* ************************************************************************* *)
(*                             TinyQCMl                                      *)
(*                   (c) Franois Pessaux - 2014                             *)
(* TinyQCMl is freely distribuable (sources and binaries if some).           *)
(* ************************************************************************* *)

(** {b Descr}: Stricly increasing version number. *)
let version_number = "1.1.0" ;;



(** {b Descr}: Flag indicating whether pdflatex must be invoked once the
    LaTeX file is generated. *)
let pdflatex_flag = ref false ;;



(** {b Descr}: Flag indicating whether an extra empty page must be generated
    to keep aligned for recto-verso print. *)
let pad_flag = ref false ;;



(** {b Descr}: Flag indicating whether item and their proposition must be
    generated in random order for each exemplary  of the MCQ. By default
    randomization is activated. *)
let random_flag = ref true ;;



(** {b Descr}: Flag indicating whether the LaTeX generated document is
    the solution of the QCM(s). Solution is generated for each exemplary. *)
let solution_flag = ref false ;;



(** {b Descr}: Name of the user TeX macros file included as preamble of the
    generated document. By default the file is "macros.tex". *)
let tex_macros_file = ref (Installation.install_lib_dir ^ "/macros.tex") ;;



(** {b Descr}: Name of the core TeX macros file included as preamble of the
    generated document. This file is intended to contain macros on which
    the generation engine hevealy relies. It may be modified only if kept
    compatible with the macros used by the engine. *)
let tex_coremacros_file = Installation.install_lib_dir ^ "/coremacros.tex" ;;



(** {b Descr}: Name of the logo to put in the banner in top of the first
    page. It default value is "logo.png" but can me overriden using the
    "-l" option. *)
let logo_file = ref (Installation.install_lib_dir ^ "/images/logo.png") ;;



(** {b Descr}: Date to display in the banner in top of the first
    page. It default value is "The day of exam" which is not meaningful and
    must be set using the "-d" option. *)
let date = ref "" ;;



(** {b Descr}: Title to display in the banner in top of the first page, next
    to the string " - QCM". Can be changed using the "-t" option. *)
let title = ref "MCQ" ;;



(** {b Descr}: Function generating the text to insert at the beginning of the
    generated LaTeX file.
    It triggers inclusion of the file 'macros.tex' (or alternative one if
    specified by the user by the '-m' option) that must be present to
    allow compilation of the LaTeX file.
    We need a function to do that and not a simple string since the name
    of the file can be changed by the '-m' option, hence we need to delay
    the evaluation of the string until options have been parsed. *)
let main_prelude () =
"\
\\documentclass[11pt,a4paper]{article}\n\
\\input{" ^ tex_coremacros_file ^ "}\n\
\\input{" ^ !tex_macros_file ^ "}\n\
\n\
\\begin{document}\n\
\n\
";;



(** {b Descr}: Preluded generated fo each exemplary of MCQ. Mostly contains
    the title, logo and name-surname fields. *)
let page_prelude () =
"\
\\hrule\\vskip1mm\n\
\\noindent\n\
\\parbox{6cm}{\\textsf{\\textbf{\\Large {" ^ !title ^ "}}}\\\\\n\
\\textsf{\\large{"^ !date ^ "}}}\\hfill\n\
\\parbox{6cm}{\\hfill\\includegraphics[scale=0.5]{" ^
     !logo_file ^ "}}\n\
\\vskip1mm\\hrule\n\
\n\
\\vskip8mm\n\
\\noindent{\\bf \\Large NOM : \\hskip6cm Prénom :}\n\
\n\
\\vskip8mm\n\
"
;;



(** {b Descr}: Text to append at the end of the generated LaTeX file.
    It mostly closes the LaTeX document. *)
let postlude =
"\
\n\
\\end{document}\n\
"
;;



exception External_cmd_error of (string * int) ;;
exception No_solution of int ;;


(** {b Descr}: Simple Knuth shuffle algorithm. Generates a random permutation
    in the interval [0 - (n-1)]. *)
let random_sequence n =
  (* Create a fresh array whose values are their indices. *)
  let perm = Array.init n (fun i -> i) in
  for i = 0 to n - 1 do
    let j = Random.int (i + 1) in
    perm.(i) <- perm.(j) ;
    perm.(j) <- i
  done ;
  (* DEBUG
  Printf.eprintf "Permutation:" ;
  Array.iter (fun v -> Printf.eprintf " %d" v) perm ;
  Printf.eprintf "\n" ;
  *)
  perm
;;



(** {b Descr}: Simple Fisher-Yates shuffle algorithm. Generates a random
    permutation in the interval [0 - (n-1)]. *)
let random_sequence2 n =
  (* Create a fresh array whose values are their indices. *)
  let perm = Array.init n (fun i -> i) in
  for _i = 0 to n - 1 do
    let j = Random.int n in
    let k = Random.int n in
    let t = perm.(j) in
    perm.(j) <- perm.(k) ;
    perm.(k) <- t
  done ;
  perm
;;



(** {b Desrc} Emit proposition for a multiple choice question. Shuffled order
    is already computed and passed as argument. *)
let emit_propositions out_hd props_order props =
  let nb_props = Array.length props in
  (* Recursively walk along the arrays, harvesting and returning the indices
     of the valid answers in the answers array of the current question. *)
  let rec emit_proposition i =
    if i = nb_props then []
    else (
      let prop = props.(props_order.(i)) in
      Printf.fprintf out_hd "{ %s }\n" prop.Parsetree.p_text ;
      (* Check for answer vaidity.
         ATTENTION: since answers are shuffled, the valid index is
         the suffled one, hence the current index in the permutation array. *)
      if prop.Parsetree.p_valid then
        i :: emit_proposition (i + 1)
      else emit_proposition (i + 1)
     ) in
  (* Now do the job. *)
  emit_proposition 0
;;



(** {b Desrc} Emit a question being a multiple choices one. Select the variant
    and output it. Shuffling of propositions is computed here but output of
    propositions is done in [emit_propositions]. *)
let emit_item out_hd item q_num q_index =
  Printf.fprintf out_hd "\\qcm{%d}{\n%s\n}\n"
    (q_num + 1) item.Parsetree.i_text ;
  let nb_props = Array.length item.Parsetree.i_props in
  (* Generate each proposition of the item in random order if random is
     requested. *)
  let props_order =
    if !random_flag then random_sequence nb_props
    else Array.init nb_props (fun i -> i) in
  (* Generate the propositions and recover indices of the valid ones. *)
  let valids =
    emit_propositions out_hd props_order item.Parsetree.i_props in
  (* Print the solution. *)
  match valids with
  | [] -> raise (No_solution q_index)
  | [one] ->
      (* Index of box to fill.
         Number items from 1, this is more human-usual. *)
      Printf.fprintf out_hd "{ %d }\n\n" (one + 1)
  | _ ->
      (* No only one solution. LaTeX macros don't handle. So do not output
         material to turn checkboxes filled. *)
      Printf.fprintf out_hd "{ }\n" ;
      Printf.fprintf out_hd ("\\ifsolution \\noindent\nCorrect :") ;
      (* Put a heading space and no space after the colon. This will avoid
         trailing space at the end of the list of correct answers. *)
      List.iter (fun i -> Printf.fprintf out_hd " %d" (i + 1)) valids ;
      Printf.fprintf out_hd (".\n\\fi\n\n")
;;



(** {b Desrc} Emit a question being a free text one. Select the variant and
    output it. *)
let emit_free_text out_hd free_text q_num =
  Printf.fprintf out_hd "\\freetext{%d}{%fcm}{\n%s\n}\n{\n%s\n}\n\n"
    (q_num + 1) free_text.Parsetree.ft_frame_height
    free_text.Parsetree.ft_text free_text.Parsetree.ft_answer
;;



(** {b Desrc} Emit the questions of the exam, mixed them randomly, using
    the lookup table of order. Each question is at this point a set of
    variants. The effective variant to use will be selected later. *)
let emit_questions out_hd questions questions_order =
  Array.iteri
    (fun q_num q_index ->
      (* Get the effective question in the questions array. *)
      match questions.(q_index) with
      | Parsetree.QChoices items ->
          (* Get randomly the effective item in those of the question. *)
          let item_index = Random.int (Array.length items) in
          let item = items.(item_index) in
          emit_item out_hd item q_num q_index
      | Parsetree.QFree free_texts ->
          (* Get randomly the effective free text in those of the question. *)
          let ft_index = Random.int (Array.length free_texts) in
          let free_text = free_texts.(ft_index) in
          emit_free_text out_hd free_text q_num
      )
    questions_order ;
  if !pad_flag then Printf.fprintf out_hd "\n\\newpage\n\\ \n" ;
;;



let main () =
  let nb_pages = ref 1 in
  let in_file_name = ref "" in
  let out_file_name = ref "qcms.tex" in
  Arg.parse
    [ ("-d", Arg.String (fun s -> date := s),
       "Set the date (a string) to display in the header of the first page.") ;
      ("-f", Arg.Clear random_flag,
       "Fixed : disable randomization of questions and propositions.") ;
      ("-l", Arg.String (fun s -> logo_file := s),
       "Set the name of the file containing the logo to display in the header \
        of \n     the first page. By default, searched in " ^
        !logo_file ^ "\n     which is a 270x200 image.") ;
      ("-m", Arg.String (fun s -> tex_macros_file := s),
      "Set the name of user TeX macros file included as preamble of the \
       document.\n     By default, searched in " ^ !tex_macros_file ^ ".") ;
      ("-n", Arg.Int (fun i -> nb_pages := i),
       "Set the number of exemplaries of the QCM. Default is 1.") ;
      ("-o", Arg.String (fun s -> out_file_name := s),
      "Set the name of the output LaTeX file.") ;
      ("-p", Arg.Set pdflatex_flag,
       "Run pdflatex once the LaTeX file is written.") ;
      ("-pad", Arg.Set pad_flag,
       "Generate an extra empty page to keep aligned for recto-verso print.") ;
      ("-s", Arg.Set solution_flag,
       "Generate the solution of the QCM instead of the ready-to-fill \
        version.") ;
      ("-t", Arg.String (fun s -> title := s),
       "Set the title to display in the header of the first page. Default is \
        empty.") ;
      ("-v", Arg.Unit
               (fun () -> Printf.printf "mkqcm v %s.\n" version_number; exit 0),
      "Print version number then exit.")
    ]
    (fun fname -> in_file_name := fname)
    "TinyQCMl v1.1\nUsage: mkqcm [options] <QCM description file>" ;
  if !in_file_name <> "" then (
    let questions = Parsefile.parse !in_file_name in
    let out_hd = open_out_bin !out_file_name in
    Printf.fprintf out_hd "%s" (main_prelude ()) ;
    (* If the solution is requested, set the LaTeX document flag. *)
    if !solution_flag then Printf.fprintf out_hd "\\solutiontrue\n" ;
    let nb_questions = Array.length questions in
    for _i = 0 to !nb_pages - 1 do
      (* Output the page prelude for each exemplary. *)
      Printf.fprintf out_hd "%s" (page_prelude ()) ;
      (* Generate the questions in random order is random is requested. *)
      let questions_order =
        if !random_flag then
          random_sequence nb_questions
        else Array.init nb_questions (fun i -> i) in
      emit_questions out_hd questions questions_order ;
      Printf.fprintf out_hd "\\newpage\n"
    done ;
    Printf.fprintf out_hd "%s" postlude ;
    close_out out_hd ;
    (* If requested, invoke pdflatex. *)
    if !pdflatex_flag then (
      Printf.printf "Invoking pdflatex on the output.\n" ;
      let cmd = "pdflatex " ^ !out_file_name in
      let ret_code = Sys.command cmd in
      if ret_code <> 0 then raise (External_cmd_error (cmd, ret_code))
    )
   )
  else (
    Printf.eprintf "Error. No input file. mkqcm doesn't know what to do.\n" ;
    exit (-1)
   )
;;



try main () with
  (* System errors. *)
  | Sys_error m ->
      Printf.eprintf "%tSystem error - %s.\n" Parsefile.pp_err_header m
  (* Lexing / parsing stage. *)
  | Parsefile.Lex_error (pos_s, pos_e, reason) ->
      Printf.eprintf
        "%a:\n%tLexical error %s.\n"
        Parsefile.pp_err_loc (pos_s, pos_e) Parsefile.pp_err_header reason
  | Parsefile.Syntax_error (pos_s, pos_e) ->
      Printf.eprintf "%a:\n%tSyntax error.\n"
        Parsefile.pp_err_loc (pos_s, pos_e) Parsefile.pp_err_header
  | Parsefile.Unclear_error (exc_string, pos_s, pos_e) ->
      Printf.eprintf
        "%a:\n%tUnclear syntax error: %s.\n"
        Parsefile.pp_err_loc (pos_s, pos_e) Parsefile.pp_err_header exc_string
  | External_cmd_error (cmd, err_code) ->
      Printf.eprintf "%tExternal command failed. Return code %d for: %s.\n"
        Parsefile.pp_err_header err_code cmd
  | No_solution index ->
      Printf.eprintf "%tItem #%d (0-based) has no valid answer.\n"
        Parsefile.pp_err_header index
  (* The ultimate firewall. *)
  | x ->
     Printf.eprintf
       "%tUnexpected error: \"%s\".\nPlease report.\n"
       Parsefile.pp_err_header (Printexc.to_string x)
;;
